@@ -1,14 +1,24 @@
 """Subclass of `Machine` that Specifically work with TPFs"""
+import os
+import logging
 import numpy as np
 import lightkurve as lk
 from astropy.coordinates import SkyCoord, match_coordinates_sky
 from astropy.time import Time
+from astropy.io import fits
 import astropy.units as u
 import matplotlib.pyplot as plt
+from matplotlib import patches
+import urllib.request
+import tarfile
 
 from .utils import get_gaia_sources
+from .aperture import estimate_source_centroids_aperture, aperture_mask_to_2d
 from .machine import Machine
+from .version import __version__
+from psfmachine import PACKAGEDIR
 
+log = logging.getLogger(__name__)
 __all__ = ["TPFMachine"]
 
 
@@ -61,22 +71,15 @@ class TPFMachine(Machine):
         )
         self.tpfs = tpfs
 
-        # Cut out 1.5 days after every data gap
-        dt = np.hstack([10, np.diff(time)])
-        focus_mask = ~np.in1d(
-            np.arange(len(time)),
-            np.hstack(
-                [
-                    np.arange(t, t + int(1.5 / np.median(dt)))
-                    for t in np.where(dt > (np.median(dt) * 5))[0]
-                ]
-            ),
-        )
-        if time_mask is None:
+        # combine focus mask and time mask if they exist
+        if time_mask is None and focus_mask is None:
+            self.time_mask = np.ones(len(self.time), bool)
+        elif time_mask is None and focus_mask is not None:
             self.time_mask = focus_mask
+        elif time_mask is not None and focus_mask is None:
+            self.time_mask = time_mask
         else:
             self.time_mask = time_mask & focus_mask
-
         self.pix2obs = pix2obs
         #        self.pos_corr1 = pos_corr1
         #        self.pos_corr2 = pos_corr2
@@ -85,15 +88,36 @@ class TPFMachine(Machine):
     def __repr__(self):
         return f"TPFMachine (N sources, N times, N pixels): {self.shape}"
 
-    def fit_lightcurves(self, plot=False, fit_va=True, iter_negative=True):
+    def fit_lightcurves(
+        self,
+        plot=False,
+        fit_va=True,
+        iter_negative=True,
+        load_shape_model=False,
+        shape_model_file=None,
+        sap=True,
+    ):
         """
         Fit the sources inside the TPFs passed to `TPFMachine`.
+        This function creates the `lcs` attribuite that contains a collection of light
+        curves in the form of `lightkurve.LightCurveCollection`. Each entry in the
+        collection is a `lightkurve.KeplerLightCurve` object with the different type
+        of photometry (SAP, PSF, and PSF velocity-aberration corrected). Also each
+        `lightkurve.KeplerLightCurve` object includes its asociated metadata.
+        The photometry can also be accessed independently from the following attribuites
+        that `fit_lightcurves` create:
+            * `ws` and `werrs` have the uncorrected PSF flux and flux errors.
+            * `ws_va` and `werrs_va` have the PSF flux and flux errors corrected by
+            velocity aberration.
+            * `sap_flux` and `sap_flux_err` have the flux and flux errors computed
+            using aperture mask.
 
         Parameters
         ----------
         plot : bool
             Whether or not to show some diagnostic plots. These can be helpful
-            for a user to see if the PRF and time dependent models are being calculated correctly.
+            for a user to see if the PRF and time dependent models are being calculated
+            correctly.
         fit_va : bool
             Whether or not to fit Velocity Aberration (which implicitly will try to fit
             other kinds of time variability). This will try to fit the "long term"
@@ -102,14 +126,27 @@ class TPFMachine(Machine):
             find you do not need this to be set to True. If you have the time, it
             is recommended to run it.
         iter_negative : bool
-            When fitting light curves, it isn't possible to force the flux to be positive.
-            As such, when we find there are light curves that deviate into negative
-            flux values, we can clip these targets out of the analysis and rerun the model.
+            When fitting light curves, it isn't possible to force the flux to be
+            positive.
+            As such, when we find there are light curves that deviate into negative flux
+            values, we can clip these targets out of the analysis and rerun the model.
             If iter_negative is True, PSFmachine will run up to 3 times, clipping out
             any negative targets each round.
+        load_shape_model : bool
+            Load PRF shape model from disk or not. Default models were computed from
+            FFI of the same channel and quarter.
+        shape_model_file : string
+            Path to PRF model file to be passed to `load_shape_model(input)`. If None,
+            then precomputed models will be download from Zenodo repo.
+        sap : boolean
+            Compute or not Simple Aperture Photometry. See
+            `Machine.compute_aperture_photometry()` for details.
         """
-
-        self.build_shape_model(plot=plot)
+        # use PRF model from FFI or create one with TPF data
+        if load_shape_model:
+            self.load_shape_model(input=shape_model_file, plot=plot)
+        else:
+            self.build_shape_model(plot=plot)
         self.build_time_model(plot=plot)
         self.fit_model(fit_va=fit_va)
         if iter_negative:
@@ -123,54 +160,15 @@ class TPFMachine(Machine):
                 idx += 1
                 if idx >= 3:
                     break
+        if sap:
+            self.compute_aperture_photometry(
+                aperture_size="optimal", target_complete=1, target_crowd=1
+            )
 
         self.lcs = []
         for idx, s in self.sources.iterrows():
-            ldx = np.where([idx in s for s in self.tpf_meta["sources"]])[0][0]
-            mission = self.tpf_meta["mission"][ldx].lower()
-            if s.tpf_id is not None:
-                if mission == "kepler":
-                    label, targetid = f"KIC {int(s.tpf_id)}", int(s.tpf_id)
-                elif mission == "tess":
-                    label, targetid = f"TIC {int(s.tpf_id)}", int(s.tpf_id)
-                elif mission in ["k2", "ktwo"]:
-                    label, targetid = f"EPIC {int(s.tpf_id)}", int(s.tpf_id)
-                else:
-                    raise ValueError(f"can not parse mission `{mission}`")
-            else:
-                label, targetid = s.designation, int(s.designation.split(" ")[-1])
 
-            meta = {
-                "ORIGIN": "PSFMACHINE",
-                "APERTURE": "PSF",
-                "LABEL": label,
-                "TARGETID": targetid,
-                "MISSION": mission,
-                "RA": s.ra,
-                "DEC": s.dec,
-                "PMRA": s.pmra / 1000,
-                "PMDEC": s.pmdec / 1000,
-                "PARALLAX": s.parallax,
-                "GMAG": s.phot_g_mean_mag,
-                "RPMAG": s.phot_rp_mean_mag,
-                "BPMAG": s.phot_bp_mean_mag,
-            }
-
-            attrs = [
-                "channel",
-                "module",
-                "ccd",
-                "camera",
-                "quarter",
-                "campaign",
-                "quarter",
-                "row",
-                "column",
-                "mission",
-            ]
-            for attr in attrs:
-                if attr in self.tpf_meta.keys():
-                    meta[attr.upper()] = self.tpf_meta[attr][ldx]
+            meta = self._make_meta_dict(idx, s, sap)
 
             if fit_va:
                 flux, flux_err = (
@@ -189,19 +187,88 @@ class TPFMachine(Machine):
                 meta=meta,
                 time_format="jd",
             )
+
             if fit_va:
-                lc["flux_NVA"] = (self.ws[:, idx]) * u.electron / u.second
-                lc["flux_err_NVA"] = (self.werrs[:, idx]) * u.electron / u.second
+                lc["psf_flux_NVA"] = (self.ws[:, idx]) * u.electron / u.second
+                lc["psf_flux_err_NVA"] = (self.werrs[:, idx]) * u.electron / u.second
+            if sap:
+                lc["sap_flux"] = (self.sap_flux[:, idx]) * u.electron / u.second
+                lc["sap_flux_err"] = (self.sap_flux_err[:, idx]) * u.electron / u.second
             self.lcs.append(lc)
-            self.lcs = lk.LightCurveCollection(self.lcs)
+        self.lcs = lk.LightCurveCollection(self.lcs)
         return
 
+    def _make_meta_dict(self, idx, s, sap):
+        """
+        Auxiliar function that creates dictionarywith metadata for a given source in
+        the catalog.
+
+        Parameters
+        ----------
+        idx : int
+            Source index.
+        s : pandas.Series
+            Row entry of the source in the source catalog.
+        sap : boolean
+            Add or not Simple Aperture Photometry metadata
+        """
+        ldx = np.where([idx in s for s in self.tpf_meta["sources"]])[0][0]
+        mission = self.tpf_meta["mission"][ldx].lower()
+        if s.tpf_id is not None:
+            if mission == "kepler":
+                label, targetid = f"KIC {int(s.tpf_id)}", int(s.tpf_id)
+            elif mission == "tess":
+                label, targetid = f"TIC {int(s.tpf_id)}", int(s.tpf_id)
+            elif mission in ["k2", "ktwo"]:
+                label, targetid = f"EPIC {int(s.tpf_id)}", int(s.tpf_id)
+            else:
+                raise ValueError(f"can not parse mission `{mission}`")
+        else:
+            label, targetid = s.designation, int(s.designation.split(" ")[-1])
+
+        meta = {
+            "ORIGIN": "PSFMACHINE",
+            "APERTURE": "PSF + SAP" if sap else "PSF",
+            "LABEL": label,
+            "TARGETID": targetid,
+            "MISSION": mission,
+            "RA": s.ra,
+            "DEC": s.dec,
+            "PMRA": s.pmra / 1000,
+            "PMDEC": s.pmdec / 1000,
+            "PARALLAX": s.parallax,
+            "GMAG": s.phot_g_mean_mag,
+            "RPMAG": s.phot_rp_mean_mag,
+            "BPMAG": s.phot_bp_mean_mag,
+            "SAP": "optimal" if sap else "None",
+            "FLFRCSAP": self.FLFRCSAP[idx] if sap else np.nan,
+            "CROWDSAP": self.CROWDSAP[idx] if sap else np.nan,
+        }
+
+        attrs = [
+            "channel",
+            "module",
+            "ccd",
+            "camera",
+            "quarter",
+            "campaign",
+            "quarter",
+            "row",
+            "column",
+            "mission",
+        ]
+        for attr in attrs:
+            if attr in self.tpf_meta.keys():
+                meta[attr.upper()] = self.tpf_meta[attr][ldx]
+        return meta
+
     def to_fits():
         """Save all the light curves to fits files."""
         raise NotImplementedError
 
     def lcs_in_tpf(self, tpf_number):
-        """Returns the light curves from a given TPF as a lightkurve.LightCurveCollection.
+        """
+        Returns the light curves from a given TPF as a lightkurve.LightCurveCollection.
 
         Parameters
         ----------
@@ -211,7 +278,7 @@ class TPFMachine(Machine):
         ldx = self.tpf_meta["sources"][tpf_number]
         return lk.LightCurveCollection([self.lcs[l] for l in ldx])
 
-    def plot_tpf(self, tdx):
+    def plot_tpf(self, tdx, sap=True):
         """
         Make a diagnostic plot of a given TPF in the stack
 
@@ -223,9 +290,11 @@ class TPFMachine(Machine):
         ----------
         tdx : int
             Index of the TPF to plot
+        sap : boolean
+            Overplot the pixel mask used for aperture photometry.
         """
         tpf = self.tpfs[tdx]
-        ax_tpf = tpf.plot(scale="log")
+        ax_tpf = tpf.plot(aperture_mask="pipeline" if sap else None)
         sources = self.sources.loc[self.tpf_meta["sources"][tdx]]
 
         img_extent = (
@@ -239,6 +308,16 @@ class TPFMachine(Machine):
         r += tpf.row
         c += tpf.column
 
+        # create 2D aperture mask for every source, for potting
+        if hasattr(self, "aperture_mask") and sap:
+            aperture_mask_2d = aperture_mask_to_2d(
+                self.tpfs,
+                self.tpf_meta["sources"],
+                self.aperture_mask,
+                self.column,
+                self.row,
+            )
+
         kdx = 0
         for sdx, s in sources.iterrows():
             if hasattr(self, "lcs"):
@@ -262,12 +341,37 @@ class TPFMachine(Machine):
                 if np.nansum(mod) == 0:
                     kdx += 1
                     continue
-                _ = plt.subplots(figsize=(10, 3))
+                _ = plt.subplots(figsize=(12, 4))
                 ax = plt.subplot2grid((1, 4), (0, 0), colspan=3)
                 lc.errorbar(ax=ax, c="k", lw=0.3, ls="-")
+                # plot SAP lc
+                if hasattr(lc, "sap_flux") and sap:
+                    lc.errorbar(
+                        column="sap_flux",
+                        ax=ax,
+                        c="tab:red",
+                        lw=0.3,
+                        ls="-",
+                        label="SAP",
+                    )
                 kdx += 1
                 ax = plt.subplot2grid((1, 4), (0, 3))
                 lk.utils.plot_image(mod, extent=img_extent, ax=ax)
+                # Overlay the aperture mask if asked
+                if sap:
+                    aperture_mask = aperture_mask_2d["%i_%i" % (tdx, sdx)]
+                    for i in range(r.shape[0]):
+                        for j in range(r.shape[1]):
+                            if aperture_mask[i, j]:
+                                rect = patches.Rectangle(
+                                    xy=(j + tpf.column - 0.5, i + tpf.row - 0.5),
+                                    width=1,
+                                    height=1,
+                                    color="tab:red",
+                                    fill=False,
+                                    hatch="//",
+                                )
+                                ax.add_patch(rect)
             col, row = tpf.wcs.all_world2pix([[s.ra, s.dec]], 0)[0]
             if (
                 (col < -3)
@@ -283,12 +387,257 @@ class TPFMachine(Machine):
                 edgecolor="k",
             )
 
+    def load_shape_model(self, input=None, plot=False):
+        """
+        Loads a PRF shape model from a FITs file.
+        Not implemented: By default this function will load PRF shapes computed from
+        FFI data (Kepler, K2, or TESS).
+
+        Parameters
+        ----------
+        input : string
+            Name of the file containing the shape parameters and weights. The file
+            has to be FITS format.
+            If None, then previously computed shape model from Kepler's FFI will be
+            download from https://zenodo.org/record/5504503/ and used as default.
+            The file download from Zenodo happens only the first time that shape models
+            of a given mission (e.g. Kepler, K2) are asked. Then, shape models for all
+            channels and quarters will be locally available for future use.
+            The file is stored in `psfmachine/src/psfmachine/data/`.
+        plot : boolean
+            Plot or not the mean model.
+        """
+        # By default we will load PRF model from FFI when this are ok.
+        if input is None:
+            input = (
+                f"{PACKAGEDIR}/data/ffi/ch{self.tpf_meta['channel'][0]:02}/"
+                f"{self.tpf_meta['mission'][0]}_ffi_shape_model_"
+                f"ch{self.tpf_meta['channel'][0]:02}_"
+                f"q{self.tpf_meta['quarter'][0]:02}.fits"
+            )
+            if not os.path.isfile(input):
+                # if file doesnt exist, download file bundle from zenodo:
+                tar_file = (
+                    f"{PACKAGEDIR}/data/"
+                    f"{self.tpf_meta['mission'][0]}_FFI_PRFmodels_v1.0.tar.gz"
+                )
+                if not os.path.isfile(tar_file):
+                    if not os.path.isdir(f"{PACKAGEDIR}/data/"):
+                        os.makedirs(f"{PACKAGEDIR}/data/")
+                    url = (
+                        f"https://zenodo.org/record/5504503/files/"
+                        f"{tar_file.split('/')[-1]}?download=1"
+                    )
+                    log.info(f"Downloading bundle files from: {url}")
+                    with urllib.request.urlopen(url) as response, open(
+                        tar_file, "wb"
+                    ) as out_file:
+                        out_file.write(response.read())
+                # unpack
+                with tarfile.open(tar_file) as f:
+                    f.extractall(f"{PACKAGEDIR}/data/ffi/")
+        # check if file exists and is the right format
+        if not os.path.isfile(input):
+            raise FileNotFoundError(f"No shape file: {input}")
+        log.info(f"Using shape model from: {input}")
+        if not input.endswith(".fits"):
+            # should use a custom exception for wrong file format
+            raise ValueError("File format not suported. Please provide a FITS file.")
+
+        # create source mask and uncontaminated pixel mask
+        self._get_source_mask()
+        self._get_uncontaminated_pixel_mask()
+
+        # open file
+        hdu = fits.open(input)
+        # check if shape parameters are for correct mission, quarter, and channel
+        if hdu[1].header["mission"] != self.tpf_meta["mission"][0]:
+            raise ValueError(
+                "Wrong shape model: file is for mission '%s',"
+                % (hdu[1].header["mission"])
+                + " it should be '%s'." % (self.tpf_meta["mission"][0])
+            )
+        if hdu[1].header["quarter"] != self.tpf_meta["quarter"][0]:
+            raise ValueError(
+                "Wrong shape model: file is for quarter %i,"
+                % (hdu[1].header["quarter"])
+                + " it should be %i." % (self.tpf_meta["quarter"][0])
+            )
+        if hdu[1].header["channel"] != self.tpf_meta["channel"][0]:
+            raise ValueError(
+                "Wrong shape model: file is for channel %i,"
+                % (hdu[1].header["channel"])
+                + " it should be %i." % (self.tpf_meta["channel"][0])
+            )
+        # load model hyperparameters and weights
+        self.n_r_knots = hdu[1].header["n_rknots"]
+        self.n_phi_knots = hdu[1].header["n_pknots"]
+        self.rmin = hdu[1].header["rmin"]
+        self.rmax = hdu[1].header["rmax"]
+        self.cut_r = hdu[1].header["cut_r"]
+        self.psf_w = hdu[1].data["psf_w"]
+        del hdu
+
+        # create mean model, but PRF shapes from FFI are in pixels! and TPFMachine
+        # work in arcseconds
+        self._get_mean_model()
+        # remove background pixels and recreate mean model
+        self._update_source_mask_remove_bkg_pixels()
+
+        if plot:
+            return self.plot_shape_model()
+        return
+
+    def save_shape_model(self, output=None):
+        """Saves the weights of a PRF fit to disk as a FITS file.
+
+        Parameters
+        ----------
+        output : str, None
+            Output file name. If None, one will be generated.
+        """
+        # asign a file name
+        if output is None:
+            output = "./%s_shape_model_ch%02i_q%02i.fits" % (
+                self.tpf_meta["mission"][0],
+                self.tpf_meta["channel"][0],
+                self.tpf_meta["quarter"][0],
+            )
+
+        # create data structure (DataFrame) to save the model params
+        table = fits.BinTableHDU.from_columns(
+            [fits.Column(name="psf_w", array=self.psf_w, format="D")]
+        )
+        # include metadata and descriptions
+        table.header["object"] = ("PRF shape", "PRF shape parameters")
+        table.header["datatype"] = ("TPF stack", "Type of data used to fit shape model")
+        table.header["origin"] = ("PSFmachine.TPFMachine", "Software of origin")
+        table.header["version"] = (__version__, "Software version")
+        table.header["mission"] = (self.tpf_meta["mission"][0], "Mission name")
+        table.header["quarter"] = (
+            self.tpf_meta["quarter"][0],
+            "Quarter of observations",
+        )
+        table.header["channel"] = (self.tpf_meta["channel"][0], "Channel output")
+        table.header["n_rknots"] = (
+            self.n_r_knots,
+            "Number of knots for spline basis in radial axis",
+        )
+        table.header["n_pknots"] = (
+            self.n_phi_knots,
+            "Number of knots for spline basis in angle axis",
+        )
+        table.header["rmin"] = (self.rmin, "Minimum value for knot spacing")
+        table.header["rmax"] = (self.rmax, "Maximum value for knot spacing")
+        table.header["cut_r"] = (
+            self.cut_r,
+            "Radial distance to remove angle dependency",
+        )
+        # spline degree is hardcoded in `_make_A_polar` implementation.
+        table.header["spln_deg"] = (3, "Degree of the spline basis")
+
+        table.writeto(output, checksum=True, overwrite=True)
+
+        return
+
+    def get_source_centroids(self, method="poscor"):
+        """
+        Compute centroids for sources in pixel coordinates.
+        It implements three different methods to calculate centroids:
+            * "aperture": computes centroids from moments `
+            Machine.estimate_source_centroids_aperture()`. This needs the aperture
+            masks to becomputed in advance with `Machine.compute_aperture_photometry`.
+            Note that for sources with partial data (i.e. near TPF edges) the this
+            method is illdefined.
+            * "poscor": uses Gaia RA and Dec coordinates converted to pixel
+            space using the TPFs WCS solution and the apply each TPF 'pos_corr'
+            correction
+            * "scene": uses Gaia coordinates again, but correction is computed from
+            the TPF scene jitter using 'Machine._get_centroids()'.
+
+        Parameters
+        ----------
+        method : string
+            What type of corrected centroid will be computed.
+            If "aperture", it creates attributes `source_centroids_[column/row]_ap`.
+            If "poscor" (default), it creates attributes
+            `source_centroids_[column/row]_poscor`.
+            If "scene", it creates attributes `source_centroids_[column/row]_scene`.
+
+            Note: "poscor" and "scene" show consistent results below 10th of a pixel.
+        """
+        # use aperture mask for moments centroids
+        if method == "aperture":
+            if not hasattr(self, "aperture_mask"):
+                raise AttributeError("No aperture masks")
+            centroids = estimate_source_centroids_aperture(
+                self.aperture_mask, self.flux, self.column, self.row
+            )
+            self.source_centroids_column_ap = centroids[0] * u.pixel
+            self.source_centroids_row_ap = centroids[1] * u.pixel
+        # source centroids using pos_corr
+        if method == "poscor":
+            # get pixel coord from catalog RA, Dec and tpf WCS solution
+            col, row = (
+                self.tpfs[0]
+                .wcs.all_world2pix(self.sources.loc[:, ["ra", "dec"]].values, 0)
+                .T
+            )
+            col += self.tpfs[0].column
+            row += self.tpfs[0].row
+
+            cadno_index = np.in1d(self.tpfs[0].time.jd, self.time)
+            self.source_centroids_column_poscor = []
+            self.source_centroids_row_poscor = []
+            for i in range(self.nsources):
+                tpf_idx = [
+                    k for k, ss in enumerate(self.tpf_meta["sources"]) if i in ss
+                ][0]
+                # apply poss_corr from TPF cadences
+                self.source_centroids_column_poscor.append(
+                    col[i] + self.tpfs[tpf_idx].pos_corr1[cadno_index]
+                )
+                self.source_centroids_row_poscor.append(
+                    row[i] + self.tpfs[tpf_idx].pos_corr2[cadno_index]
+                )
+            self.source_centroids_column_poscor = (
+                np.array(self.source_centroids_column_poscor) * u.pixel
+            )
+            self.source_centroids_row_poscor = (
+                np.array(self.source_centroids_row_poscor) * u.pixel
+            )
+        # use gaia coordinates and scene centroids
+        if method == "scene":
+            if not hasattr(self, "ra_centroid"):
+                self._get_source_mask()
+            centr_ra = np.tile(self.sources.ra.values, (self.nt, 1)).T + np.tile(
+                self.ra_centroid.value, (self.nsources, 1)
+            )
+            centr_dec = np.tile(self.sources.dec.values, (self.nt, 1)).T + np.tile(
+                self.dec_centroid.value, (self.nsources, 1)
+            )
+
+            self.source_centroids_column_scene, self.source_centroids_row_scene = (
+                self.tpfs[0]
+                .wcs.all_world2pix(np.array([centr_ra.ravel(), centr_dec.ravel()]).T, 0)
+                .T
+            )
+            self.source_centroids_column_scene = (
+                self.source_centroids_column_scene.reshape(self.nsources, self.nt)
+                + self.tpfs[0].column
+            ) * u.pixel
+            self.source_centroids_row_scene = (
+                self.source_centroids_row_scene.reshape(self.nsources, self.nt)
+                + self.tpfs[0].row
+            ) * u.pixel
+
     @staticmethod
     def from_TPFs(
         tpfs,
         magnitude_limit=18,
         dr=2,
         time_mask=None,
+        apply_focus_mask=True,
         query_ra=None,
         query_dec=None,
         query_rad=None,
@@ -310,6 +659,25 @@ class TPFMachine(Machine):
         ----------
         tpfs: lightkurve TargetPixelFileCollection
             Collection of Target Pixel files
+        magnitude_limit : float
+            Limiting magnitude to query Gaia catalog.
+        dr : int
+            Gaia data release to be use, default is 2, options are DR2 and EDR3
+        time_mask : boolean array
+            Mask to be applied to discard cadences if needed.
+        apply_focus_mask : boolean
+            Mask or not cadances near observation gaps to remove focus change.
+        query_ra : numpy.array
+            Array of RA to query Gaia catalog. Default is `None` and will use the
+            coordinate centers of each TPF.
+        query_dec : numpy.array
+            Array of Dec to query Gaia catalog. Default is `None` and will use the
+            coordinate centers of each TPF.
+        query_rad : numpy.array
+            Array of radius to query Gaia catalog. Default is `None` and will use
+            the coordinate centers of each TPF.
+        **kwargs
+            Keyword arguments to be passed to `TPFMachine`.
 
         Returns
         -------
@@ -319,7 +687,8 @@ class TPFMachine(Machine):
         if not isinstance(tpfs, lk.collections.TargetPixelFileCollection):
             raise TypeError("<tpfs> must be a of class Target Pixel Collection")
 
-        # CH: all these internal functions should be put in another and from_tpfs should be in another helper module
+        # CH: all these internal functions should be put in another and from_tpfs
+        # should be in another helper module
         attrs = [
             "ra",
             "dec",
@@ -343,9 +712,9 @@ class TPFMachine(Machine):
 
         tpf_meta = {k: m for m, k in zip(meta, attrs)}
         if isinstance(tpfs[0], lk.KeplerTargetPixelFile):
-            tpf_meta["tpfmag"] = [tpf.header["kepmag"] for tpf in tpfs]
+            tpf_meta["tpfmag"] = [tpf.get_header()["kepmag"] for tpf in tpfs]
         elif isinstance(tpfs[0], lk.TessTargetPixelFile):
-            tpf_meta["tpfmag"] = [tpf.header["tmag"] for tpf in tpfs]
+            tpf_meta["tpfmag"] = [tpf.get_header()["tmag"] for tpf in tpfs]
         else:
             raise ValueError("TPFs not understood")
 
@@ -371,6 +740,8 @@ class TPFMachine(Machine):
 
         if time_mask is not None:
             time_mask = np.copy(time_mask)[qual_mask]
+        # use or not the focus mask
+        focus_mask = focus_mask if apply_focus_mask else None
 
         # convert to RA Dec
         locs, ra, dec = _wcs_from_tpfs(tpfs)
@@ -495,6 +866,7 @@ def _parse_TPFs(tpfs, **kwargs):
             tpfs[0].quality, 1 | 2 | 4 | 8 | 32 | 16384 | 65536 | 1048576
         )
         qual_mask &= (np.abs(tpfs[0].pos_corr1) < 5) & (np.abs(tpfs[0].pos_corr2) < 5)
+        # Cut out 1.5 days after every data gap
         dt = np.hstack([10, np.diff(time)])
         focus_mask = ~np.in1d(
             np.arange(len(time)),
@@ -681,26 +1053,23 @@ def _wcs_from_tpfs(tpfs):
     dec: numpy.ndarray
         Array with declination values per pixel
     """
-    # calculate x,y grid of each pixel
-    locs = np.hstack(
-        [
-            np.mgrid[
-                tpf.column : tpf.column + tpf.shape[2],
-                tpf.row : tpf.row + tpf.shape[1],
-            ].reshape(2, np.product(tpf.shape[1:]))
-            for tpf in tpfs
-        ]
-    )
+    # calculate x,y grid of each pixel and RA, Dec
+    locs, ra, dec = [], [], []
+    for tpf in tpfs:
+        ij = np.mgrid[
+            tpf.column : tpf.column + tpf.shape[2],
+            tpf.row : tpf.row + tpf.shape[1],
+        ].reshape(2, np.product(tpf.shape[1:]))
+        ra_, dec_ = tpf.wcs.wcs_pix2world(
+            np.vstack([(ij[0] - tpf.column), (ij[1] - tpf.row)]).T, 0.0
+        ).T
+        locs.append(ij)
+        ra.extend(ra_)
+        dec.extend(dec_)
+    locs = np.hstack(locs)
+    ra = np.array(ra)
+    dec = np.array(dec)
 
-    # convert pixel coord to ra, dec using TPF's solution
-    ra, dec = (
-        tpfs[0]
-        .wcs.wcs_pix2world(
-            np.vstack([(locs[0] - tpfs[0].column), (locs[1] - tpfs[0].row)]).T,
-            0.0,
-        )
-        .T
-    )
     return locs, ra, dec
 
 
@@ -736,13 +1105,13 @@ def _get_coord_and_query_gaia(
     if (ra is None) & (dec is None) & (rad is None):
         ras1, decs1 = np.asarray(
             [
-                tpf.wcs.all_pix2world([np.asarray(tpf.shape[1:]) + 4], 0)[0]
+                tpf.wcs.all_pix2world([np.asarray(tpf.shape[::-1][:2]) + 4], 0)[0]
                 for tpf in tpfs
             ]
         ).T
         ras, decs = np.asarray(
             [
-                tpf.wcs.all_pix2world([np.asarray(tpf.shape[1:]) // 2], 0)[0]
+                tpf.wcs.all_pix2world([np.asarray(tpf.shape[::-1][:2]) // 2], 0)[0]
                 for tpf in tpfs
             ]
         ).T
@@ -810,9 +1179,8 @@ def _clean_source_list(sources, ra, dec):
 
     # Keep track of sources that we removed
     sources.loc[:, "clean_flag"] = 0
-    sources.loc[:, "clean_flag"][~inside] = 2 ** 0  # outside TPF
-    # sources.loc[:, "clean_flag"][unresolved] += 2 ** 1  # close contaminant
-
+    # to avoid pandas SettingWithCopyWarning
+    sources.clean_flag.where(inside, 2 ** 0, inplace=True)
     # combine 2 source masks
     clean = sources.clean_flag == 0
     removed_sources = sources[~clean].reset_index(drop=True)
